/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.maven.shared.utils.exceptionutils;

/**
 *
 * @author <a href="mailto:struberg@yahoo.de">Mark Struberg</a>
 */
public class TestException extends Exception {
    private Throwable cause;
    private Throwable specialCause;

    public TestException() {
        super();
    }

    public void setSourceException(Throwable cause) {
        this.cause = cause;
    }

    public Throwable getSourceException() {
        return cause;
    }

    public Throwable getSpecialCause() {
        return specialCause;
    }

    public void setSpecialCause(Throwable specialCause) {
        this.specialCause = specialCause;
    }
}
